"""Implementation of ThemedApp."""

import asyncio
import json
import logging
import os
import sys
from asyncio import create_task
from functools import partial
from pathlib import Path
from typing import Callable, Optional
from warnings import warn

import blinker
import sass
from mergedeep import Strategy, merge
from trame.app import get_server
from trame.assets.local import LocalFileManager
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import client
from trame.widgets import vuetify3 as vuetify
from trame_client.widgets import html
from trame_server.core import Server
from trame_server.state import State

from nova.common.signals import Signal
from nova.mvvm.pydantic_utils import validate_pydantic_parameter
from nova.trame.view.theme.exit_button import ExitButton
from nova.trame.view.utilities.local_storage import LocalStorageManager

THEME_PATH = Path(__file__).parent

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class ThemedApp:
    """Automatically injects theming into your Trame application.

    You should always inherit from this class when you define your Trame application.
    """

    def __init__(
        self, layout: str = "default", server: Server = None, vuetify_config_overrides: Optional[dict] = None
    ) -> None:
        """Constructor for the ThemedApp class.

        Parameters
        ----------
        layout : str
            The layout to use. Current options are: :code:`default` and :code:`two-column`
        server : `trame_server.core.Server \
            <https://trame.readthedocs.io/en/latest/core.server.html#trame_server.core.Server>`_, optional
            The Trame server to use. If not provided, a new server will be created.
        vuetify_config_overrides : dict, optional
            `Vuetify Configuration <https://vuetifyjs.com/en/features/global-configuration/>`__
            that will override anything set in our default configuration. You should only use this if you don't want to
            use one of our predefined themes. If you just want to set your color palette without providing a full
            Vuetify configuration, then you can set use the following to only set the color palette used by our
            :code:`ModernTheme`:

            .. code-block:: json

                {
                    "primary": "#f00",
                    "secondary": "#0f0",
                    "accent": "#00f",
                }

        Returns
        -------
        None
        """
        self.server = get_server(server, client_type="vue3")
        self.local_storage: Optional[LocalStorageManager] = None
        self._download_file: Optional[Callable] = None
        if vuetify_config_overrides is None:
            vuetify_config_overrides = {}

        self.css = None
        try:
            with open(THEME_PATH / "assets" / "core_style.scss", "r") as scss_file:
                self.css = sass.compile(string=scss_file.read())
        except Exception as e:
            logger.warning("Could not load base scss stylesheet.")
            logger.error(e)

        theme_path = THEME_PATH / "assets" / "vuetify_config.json"
        try:
            with open(theme_path, "r") as vuetify_config:
                self.vuetify_config = json.load(vuetify_config)

                merge(
                    self.vuetify_config,
                    vuetify_config_overrides,
                    strategy=Strategy.REPLACE,
                )
        except Exception as e:
            logger.warning(f"Could not load vuetify config from {theme_path}.")
            logger.error(e)
        for shortcut in ["primary", "secondary", "accent"]:
            if shortcut in self.vuetify_config:
                self.vuetify_config["theme"]["themes"]["ModernTheme"]["colors"][shortcut] = self.vuetify_config[
                    shortcut
                ]

        self.init_lodash()

        # Since this is only intended for theming Trame apps, I don't think we need to invoke the MVVM framework here,
        # and working directly with the Trame state makes this easier for me to manage.
        self.state.nova__menu = False
        self.state.nova__defaults = self.vuetify_config["theme"]["themes"]["ModernTheme"].get("defaults", {})
        self.state.nova__theme = "ModernTheme"
        self.state.trame__favicon = LocalFileManager(__file__).url("favicon", "./assets/favicon.png")

    @property
    def state(self) -> State:
        return self.server.state

    def download_file(self, filename: str, mimetype: str, content: bytes) -> None:
        """Attempts to download a file via the browser to the user's computer.

        Note that this will do nothing if no client is connected to the Trame application when called.

        Parameters
        ----------
        filename : str
            The name of the file to be downloaded.
        mimetype : str
            The `MIME type <https://developer.mozilla.org/en-US/docs/Web/HTTP/Guides/MIME_types>`__ of the content.
        content : bytes
            The bytes to be included in the download.

        Returns
        -------
        None
        """
        if not self._download_file:
            warn(
                "You must call create_ui on this instance before download_file can be used.",
                stacklevel=1,
            )

            return

        self._download_file([filename, mimetype, content])

    def init_lodash(self) -> None:
        js_path = (Path(__file__).parent / "assets" / "js").resolve()
        self.server.enable_module(
            {
                "scripts": ["assets/js/lodash.min.js", "assets/js/delay_manager.js"],
                "serve": {"assets/js": js_path},
            }
        )

    def init_mantid(self) -> None:
        """Initializes MantidManager.

        This doesn't happen by default because Mantid is a large dependency.
        """
        pass

    async def _init_theme(self) -> None:
        if self.local_storage:
            theme = await self.local_storage.get("nova__theme")
            if theme and theme in self.vuetify_config["theme"]["themes"]:
                self.set_theme(theme, False)

    async def init_theme(self) -> None:
        create_task(self._init_theme())

    async def get_jobs_callback(self) -> None:
        get_tools_signal = blinker.signal(Signal.GET_ALL_TOOLS)
        response = get_tools_signal.send()
        if response and len(response[0]) > 1:  # Make sure that the callback had a return value
            try:
                self.server.state.nova_running_jobs = [tool.id for tool in response[0][1]]
                if len(self.server.state.nova_running_jobs) > 0:
                    self.server.state.nova_show_stop_jobs_on_exit_checkbox = True
                    self.server.state.nova_kill_jobs_on_exit = True
                else:
                    self.server.state.nova_show_stop_jobs_on_exit_checkbox = False
            except Exception as e:
                logger.warning(f"Issue getting running jobs: {e}")

    async def exit_callback(self) -> None:
        logger.info(f"Closing App. Killing jobs: {self.server.state.nova_kill_jobs_on_exit}")
        if self.server.state.nova_kill_jobs_on_exit:
            self.server.state.nova_show_exit_progress = True
            await asyncio.sleep(2)
            stop_signal = blinker.signal(Signal.EXIT_SIGNAL)
            stop_signal.send()
        sys.exit(0)

    def set_theme(self, theme: Optional[str], force: bool = True) -> None:
        """Sets the theme of the application.

        Parameters
        ----------
        theme : str, optional
            The new theme to use. If the theme is not found, the default theme will be used. The available options are:

            1. ModernTheme (default) - Leverages ORNL brand colors and a typical Vuetify appearance.
            2. CompactTheme - Similar to ModernTheme but with a smaller global font size and increased density.
        force : bool, optional
            If True, the theme will be set even if the theme selection menu is disabled.

        Returns
        -------
        None
        """
        if theme not in self.vuetify_config["theme"]["themes"]:
            raise ValueError(
                f"Theme '{theme}' not found in the Vuetify configuration. "
                "For a list of available themes, please visit "
                "https://nova-application-development.readthedocs.io/en/stable/api.html#nova.trame.ThemedApp."
            )

        # I set force to True by default as I want the user to be able to say self.set_theme('MyTheme')
        # while still blocking theme.py calls to set_theme if the selection menu is disabled.
        if self.state.nova__menu or force:
            with self.state:
                self.state.nova__defaults = self.vuetify_config["theme"]["themes"].get(theme, {}).get("defaults", {})
                self.state.nova__theme = theme

        # We only want to sync to localStorage if the user is selecting and we want to preserve the selection.
        if self.state.nova__menu and self.local_storage:
            self.local_storage.set("nova__theme", theme)

    def create_ui(self) -> VAppLayout:
        """Creates the base UI into which you will inject your content.

        You should always call this method from your application class that inherits from :code:`ThemedApp`.

        Returns
        -------
        `trame_client.ui.core.AbstractLayout <https://trame.readthedocs.io/en/latest/core.ui.html#trame_client.ui.core.AbstractLayout>`_
        """
        # This detects if Pixi is running in production mode so that we can show links to NOVA resources during
        # development.
        show_nova_resources = os.environ.get("PIXI_ENVIRONMENT_NAME", "") != "production"

        with VAppLayout(self.server, vuetify_config=self.vuetify_config) as layout:
            self.local_storage = LocalStorageManager(self.server.controller)
            self._download_file = client.JSEval(
                exec=(
                    "async ($event) => {"
                    "  const [filename, mimetype, content] = $event;"
                    "  const blob = new window.Blob([content], {type: mimetype});"
                    "  const url = window.URL.createObjectURL(blob);"
                    "  const anchor = window.document.createElement('a');"
                    "  anchor.setAttribute('href', url);"
                    "  anchor.setAttribute('download', filename);"
                    "  window.document.body.appendChild(anchor);"
                    "  anchor.click();"
                    "  window.document.body.removeChild(anchor);"
                    "  window.setTimeout(() => window.URL.revokeObjectURL(url), 1000);"
                    "}"
                )
            ).exec

            client.ClientTriggers(mounted=self.init_theme)
            client.Style(self.css)

            with vuetify.VDefaultsProvider(defaults=("nova__defaults",)) as defaults:
                layout.defaults = defaults

                with vuetify.VThemeProvider(theme=("nova__theme",)) as theme:
                    layout.theme = theme

                    with vuetify.VAppBar() as toolbar:
                        layout.toolbar = toolbar

                        with vuetify.VAppBarTitle(classes="flex-0-1") as toolbar_title:
                            layout.toolbar_title = toolbar_title

                        if show_nova_resources:
                            vuetify.VBtn(
                                "NOVA Examples",
                                classes="ml-4",
                                href="https://github.com/nova-model/nova-examples/",
                                __properties=["target"],
                                target="_blank",
                            )
                            html.Div("·", classes="mx-1")
                            vuetify.VBtn(
                                "NOVA Tutorial",
                                href="https://nova.ornl.gov/tutorial/",
                                __properties=["target"],
                                target="_blank",
                            )
                            html.Div("·", classes="mx-1")
                            vuetify.VBtn(
                                "NOVA Documentation",
                                href="https://nova-application-development.readthedocs.io/en/latest/",
                                __properties=["target"],
                                target="_blank",
                            )

                        vuetify.VSpacer()
                        with html.Div(classes="mr-2") as actions:
                            layout.actions = actions

                            with vuetify.VMenu(
                                v_if="nova__menu",
                                close_delay=10000,
                                open_on_hover=True,
                            ) as theme_menu:
                                layout.theme_menu = theme_menu

                                with vuetify.Template(v_slot_activator="{ props }"):
                                    vuetify.VBtn(
                                        v_bind="props",
                                        classes="mr-2",
                                        icon="mdi-brush-variant",
                                    )

                                with vuetify.VList(width=200):
                                    vuetify.VListSubheader("Select Theme")

                                    for theme in self.vuetify_config.get("theme", {}).get("themes", {}).values():
                                        with vuetify.VListItem(click=partial(self.set_theme, theme["value"])):
                                            vuetify.VListItemTitle(theme["title"])
                                            vuetify.VListItemSubtitle(
                                                "Selected",
                                                v_if=f"nova__theme === '{theme['value']}'",
                                            )
                            ExitButton(self.exit_callback, self.get_jobs_callback)

                    with vuetify.VMain(classes="align-stretch d-flex flex-column h-screen"):
                        # [slot override example]
                        layout.pre_content = vuetify.VSheet(classes="bg-background flex-0-1 mt-1 nova-pre-content ")
                        # [slot override example complete]
                        with vuetify.VContainer(classes="flex-1-1 overflow-hidden pt-0 pb-2", fluid=True):
                            layout.content = vuetify.VCard(
                                classes="d-flex flex-column flex-1-1 h-100 overflow-y-auto pa-1 "
                            )
                        layout.post_content = vuetify.VSheet(classes="bg-background flex-0-1 mb-1 ")

                    with vuetify.VFooter(
                        app=True,
                        classes="my-0 px-1 py-0 text-center justify-center",
                        border=True,
                    ) as footer:
                        layout.footer = footer

                        vuetify.VProgressCircular(
                            classes="mr-1",
                            color="primary",
                            indeterminate=("!!galaxy_running",),
                            size=16,
                            width=3,
                        )
                        html.A(
                            "Powered by Calvera",
                            classes="text-grey-lighten-1 text-caption text-decoration-none",
                            href=("galaxy_url",),
                            target="_blank",
                        )
                        vuetify.VSpacer()
                        footer.add_child(
                            '<a href="https://www.ornl.gov/" '
                            'class="text-grey-lighten-1 text-caption text-decoration-none" '
                            'target="_blank">© {{ new Date().getFullYear() }} ORNL</a>'
                        )

            @self.server.controller.trigger("validate_pydantic_field")
            def validate_pydantic_field(name: str, value: str, index: int) -> bool:
                if "[index]" in name:
                    name = name.replace("[index]", f"[{str(index)}]")
                return validate_pydantic_parameter(name, value)

            return layout



"""View model for the config model."""

from typing import Optional

from nova.mvvm.interface import BindingInterface

from ..models.config import Config, LocalStorageState


class ConfigVM:
    """View model for the config model."""

    def __init__(self, binding: BindingInterface) -> None:
        self.config = Config()
        self.local_storage_state = LocalStorageState()

        self.config_bind = binding.new_bind(self.config)
        self.local_storage_bind = binding.new_bind(self.local_storage_state)

    def append_to_autoscroll(self, value: str) -> None:
        self.config.autoscroll += value
        self.config_bind.update_in_view(self.config)

    def get_local_storage(self) -> str:
        return self.local_storage_state.value

    def set_local_storage(self, value: Optional[str]) -> None:
        if value is not None:
            self.local_storage_state.value = value
            self.local_storage_bind.update_in_view(self.local_storage_state)



"""View model for RemoteFileInput."""

import os
from typing import Any, Dict, List, Union

from nova.mvvm.interface import BindingInterface
from nova.trame.model.remote_file_input import RemoteFileInputModel


class RemoteFileInputViewModel:
    """Manages the view state of RemoteFileInput."""

    counter = 0

    def __init__(self, model: RemoteFileInputModel, binding: BindingInterface) -> None:
        """Creates a new RemoteFileInputViewModel."""
        self.model = model
        self.binding = binding

        # Needed to keep state variables separated if this class is instantiated multiple times.
        self.id = RemoteFileInputViewModel.counter
        RemoteFileInputViewModel.counter += 1

        self.showing_all_files = False
        self.showing_base_paths = True
        self.previous_value = ""
        self.value = ""
        self.dialog_bind = self.binding.new_bind()
        self.file_list_bind = self.binding.new_bind()
        self.filter_bind = self.binding.new_bind()
        self.showing_all_bind = self.binding.new_bind()
        self.valid_selection_bind = self.binding.new_bind()
        self.on_close_bind = self.binding.new_bind()
        self.on_update_bind = self.binding.new_bind()

    def reset_update_binding(self) -> None:
        self.on_update_bind = self.binding.new_bind()

    def open_dialog(self) -> None:
        self.previous_value = self.value
        self.populate_file_list()

        self.dialog_bind.update_in_view(True)

    def close_dialog(self, cancel: bool = False) -> None:
        if not cancel:
            self.on_update_bind.update_in_view(self.value)
        else:
            self.value = self.previous_value

        self.filter_bind.update_in_view(self.value)
        self.on_close_bind.update_in_view(None)

    def filter_paths(self, filter: str) -> None:
        if not filter:
            self.select_file("")
        self.populate_file_list(filter)

    def get_dialog_state_name(self) -> str:
        return f"nova__dialog_{self.id}"

    def get_file_list_state_name(self) -> str:
        return f"nova__file_list_{self.id}"

    def get_filter_state_name(self) -> str:
        return f"nova__filter_{self.id}"

    def get_showing_all_state_name(self) -> str:
        return f"nova__showing_all_{self.id}"

    def get_valid_selection_state_name(self) -> str:
        return f"nova__valid_selection_{self.id}"

    def init_view(self) -> None:
        self.dialog_bind.update_in_view(False)
        self.valid_selection_bind.update_in_view(False)
        self.showing_all_bind.update_in_view(self.showing_all_files)

    def set_value(self, value: str) -> None:
        self.value = value

    def toggle_showing_all_files(self) -> None:
        self.showing_all_files = not self.showing_all_files
        self.showing_all_bind.update_in_view(self.showing_all_files)
        self.populate_file_list()

    def populate_file_list(self, filter: str = "") -> None:
        files = self.scan_current_path(filter)
        self.file_list_bind.update_in_view(files)
        if filter:
            parent = os.path.dirname(self.value)
            if os.path.isfile(self.value):
                parent = os.path.dirname(parent)

            if filter.rstrip("/") == parent:
                self.select_file(filter)
            else:
                for file in files:
                    if filter == file["path"] or filter == f"{self.value}/{file['path']}":
                        self.select_file(filter)

    def scan_current_path(self, filter: str) -> List[Dict[str, Any]]:
        files, self.showing_base_paths = self.model.scan_current_path(self.value, self.showing_all_files, filter)

        return files

    def select_file(self, file: Union[dict[str, str], str]) -> None:
        new_path = self.model.select_file(file, self.value, self.showing_base_paths)
        self.set_value(new_path)
        self.filter_bind.update_in_view(self.value)

        self.valid_selection_bind.update_in_view(self.model.valid_selection(new_path))
        self.populate_file_list()

    def set_binding_parameters(self, **kwargs: Any) -> None:
        self.model.set_binding_parameters(**kwargs)



"""Unit tests for LocalStorageManager."""

from selenium.webdriver import ActionChains, Firefox
from selenium.webdriver.common.by import By


def test_tool_components(driver: Firefox) -> None:
    run_button = driver.find_element(By.ID, "execution_test_run")
    cancel_button = driver.find_element(By.ID, "execution_test_cancel")
    progress_bar_running = driver.find_element(By.ID, "progress_bar_test_show_progress")
    progress_bar_finished = driver.find_element(By.ID, "progress_bar_test_show_ok")
    outputs = driver.find_element(By.ID, "tool_outputs_test_outputs")
    errors = driver.find_element(By.ID, "tool_outputs_test_errors")

    ActionChains(driver).click(run_button).pause(2).perform()
    # The progress bar should be visible and the output textareas should have content.
    running_style = progress_bar_running.get_attribute("style")
    assert running_style and "display: none;" not in running_style
    assert outputs.get_attribute("value") == "test_output"
    assert errors.get_attribute("value") == "test_error"

    ActionChains(driver).click(cancel_button).pause(2).perform()
    # The finished bar should be visible.
    running_style = progress_bar_running.get_attribute("style")
    assert running_style and "display: none;" in running_style
    finished_style = progress_bar_finished.get_attribute("style")
    assert finished_style and "display: none;" not in finished_style



"""Module for the Tool ouputs ViewModel."""

from typing import Any

import blinker

from nova.common.job import ToolOutputs
from nova.common.signals import Signal, get_signal_id
from nova.mvvm.interface import BindingInterface


class ToolOutputsViewModel:
    """A viewmodel responsible for tool stdout and stderr."""

    def __init__(self, id: str, binding: BindingInterface):
        self.tool_outputs = ToolOutputs()
        self.tool_outputs_bind = binding.new_bind(self.tool_outputs)
        self.outputs_signal = blinker.signal(get_signal_id(id, Signal.OUTPUTS))
        self.outputs_signal.connect(self.on_outputs_update, weak=False)

    async def on_outputs_update(self, _sender: Any, outputs: ToolOutputs) -> None:
        self.tool_outputs = outputs
        self.tool_outputs_bind.update_in_view(self.tool_outputs)



"""Tests theme.py."""

from nova.trame import ThemedApp


def test_theme_configuration() -> None:
    app = ThemedApp(vuetify_config_overrides={"primary": "#ff0000"})
    assert app.vuetify_config["theme"]["themes"]["ModernTheme"]["colors"]["primary"] == "#ff0000"


def test_set_theme() -> None:
    app = ThemedApp()
    assert app.state.nova__theme == "ModernTheme"
    app.set_theme("CompactTheme", force=True)
    assert app.state.nova__theme == "CompactTheme"
    try:
        app.set_theme("FakeTheme")
        raise AssertionError("Expected ValueError")
    except ValueError as e:
        assert str(e).startswith("Theme 'FakeTheme' not found")


def test_inheritance() -> None:
    # [setup app]
    from trame.widgets import vuetify3 as vuetify
    from trame_server.core import Server

    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            super().__init__(server=server)

            self.create_ui()

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    vuetify.VBtn("Hello World")

    # [setup app complete]



"""View model for MVVM demo of DataSelector."""

from typing import Any, Dict

from nova.mvvm.interface import BindingInterface
from tests.gallery.models.data_selector import DataSelectorBindingTest, DataSelectorState


class DataSelectorVM:
    """View model for MVVM demo of DataSelector."""

    def __init__(self, binding: BindingInterface) -> None:
        self.model = DataSelectorState()
        self.binding_test = DataSelectorBindingTest()

        self.model_bind = binding.new_bind(self.model, callback_after_update=self.on_update)
        self.parameter_bind = binding.new_bind(self.binding_test, callback_after_update=self.on_parameter_update)

    def on_update(self, data: Dict[str, Any]) -> None:
        print(data, self.model)

    def on_parameter_update(self, data: Dict[str, Any]) -> None:
        print(data, self.binding_test)



"""View Implementation for DataSelector."""

from asyncio import ensure_future, sleep
from typing import Any, List, Tuple, Union
from warnings import warn

from trame.app import get_server
from trame.widgets import client, html
from trame.widgets import vuetify3 as vuetify
from trame_server.core import State

from nova.mvvm._internal.utils import rgetdictvalue
from nova.mvvm.trame_binding import TrameBinding
from nova.trame._internal.utils import get_state_param, set_state_param
from nova.trame.model.data_selector import DataSelectorModel, DataSelectorState
from nova.trame.utils.types import TrameTuple
from nova.trame.view.layouts import GridLayout, HBoxLayout, VBoxLayout
from nova.trame.view_model.data_selector import DataSelectorViewModel

from .input_field import InputField

vuetify.enable_lab()


class DataSelector(vuetify.VDataTableVirtual):
    """Allows the user to select datafiles from the server."""

    def __init__(
        self,
        v_model: Union[str, Tuple],
        directory: Union[str, Tuple],
        clear_selection_on_directory_change: Union[bool, Tuple] = True,
        extensions: Union[List[str], Tuple, None] = None,
        prefix: Union[str, Tuple] = "",
        subdirectory: Union[str, Tuple] = "",
        refresh_rate: Union[int, Tuple] = 30,
        select_strategy: Union[str, Tuple] = "all",
        show_selected_files: Union[bool, Tuple] = True,
        **kwargs: Any,
    ) -> None:
        """Constructor for DataSelector.

        For all parameters, tuples have a special syntax. See :ref:`TrameTuple <api_trame_tuple>` for a description of
        it.

        Parameters
        ----------
        v_model : Union[str, Tuple]
            The name of the state variable to bind to this widget. The state variable will contain a list of the files
            selected by the user.
        directory : Union[str, Tuple]
            The top-level folder to expose to users. Only contents of this directory and its children will be exposed to
            users.
        clear_selection_on_directory_change: Union[bool, Tuple], optional
            Whether or not to clear the selected files when the directory is changed.
        extensions : Union[List[str], Tuple], optional
            A list of file extensions to restrict selection to. If unset, then all files will be shown.
        prefix : Union[str, Tuple], optional
            Deprecated. Please refer to the `subdirectory` parameter.
        subdirectory : Union[str, Tuple], optional
            A subdirectory within the selected top-level folder to show files. If not specified as a string, the user
            will be shown a folder browser and will be able to see all files in the selected top-level folder.
        refresh_rate : Union[int, Tuple], optional
            The number of seconds between attempts to automatically refresh the file list. Set to zero to disable this
            feature. Defaults to 30 seconds.
        select_strategy : Union[str, Tuple], optional
            The selection strategy to pass to the `VDataTable component <https://trame.readthedocs.io/en/latest/trame.widgets.vuetify3.html#trame.widgets.vuetify3.VDataTable>`__.
            If unset, the `all` strategy will be used.
        show_selected_files : Union[bool, Tuple], optional
            If true, then the currently selected files will be shown to the user below the directory and file selection
            widgets.
        **kwargs
            All other arguments will be passed to the underlying
            `VDataTable component <https://trame.readthedocs.io/en/latest/trame.widgets.vuetify3.html#trame.widgets.vuetify3.VDataTable>`_.

        Returns
        -------
        None
        """
        if "allow_custom_directory" in kwargs or "facility" in kwargs or "instrument" in kwargs:
            raise TypeError(
                "The old DataSelector component has been renamed to NeutronDataSelector. Please import it from "
                "`nova.trame.view.components.ornl`."
            )

        if "items" in kwargs:
            raise AttributeError("The items parameter is not allowed on DataSelector widget.")

        if "label" in kwargs:
            self._label = kwargs["label"]
        else:
            self._label = None

        if prefix:
            warn(
                "The prefix parameter has been deprecated. Please switch to using the subdirectory parameter.",
                category=DeprecationWarning,
                stacklevel=1,
            )

            if not subdirectory:
                subdirectory = prefix

        self._v_model = v_model
        if isinstance(v_model, str):
            self._v_model_name_in_state = v_model.split(".")[0]
        else:
            self._v_model_name_in_state = v_model[0].split(".")[0]

        self._clear_selection = clear_selection_on_directory_change
        self._directory = directory
        self._last_directory = get_state_param(self.state, self._directory)
        self._extensions = extensions if extensions is not None else []
        self._last_extensions = get_state_param(self.state, self._extensions)
        self._subdirectory = subdirectory
        self._last_subdirectory = get_state_param(self.state, self._subdirectory)
        self._refresh_rate = refresh_rate
        self._select_strategy = select_strategy
        self._show_selected_files = TrameTuple.create(show_selected_files)

        self._state_name = f"nova__dataselector_{self._next_id}_state"
        self._directories_name = f"nova__dataselector_{self._next_id}_directories"
        self._datafiles_name = f"nova__dataselector_{self._next_id}_datafiles"

        self._flush_state = f"flushState('{self._v_model_name_in_state}');"
        self._reset_state = client.JSEval(exec=f"{self._v_model} = []; {self._flush_state}").exec

        self._create_model()
        self._create_viewmodel()
        self._setup_bindings()

        self.create_ui(**kwargs)

        ensure_future(self._refresh_loop())

    @property
    def state(self) -> State:
        return get_server(None, client_type="vue3").state

    def create_ui(self, *args: Any, **kwargs: Any) -> None:
        show_directories = isinstance(self._subdirectory, tuple) or not self._subdirectory

        with VBoxLayout(classes="nova-data-selector", stretch=True) as self._layout:
            with HBoxLayout(valign="center"):
                self._layout.filter = html.Div(classes="flex-1-1")
                with vuetify.VBtn(icon=True, size="small", variant="text"):
                    vuetify.VIcon("mdi-magnify", size=16)
                    vuetify.VTooltip("Search", activator="parent")

                    with vuetify.VMenu(activator="parent", close_on_content_click=False):
                        with vuetify.VCard(width=200):
                            with vuetify.VCardText():
                                InputField(v_model=f"{self._state_name}.search", variant="outlined")

                with vuetify.VBtn(icon=True, size="small", variant="text", click=self.refresh_contents):
                    vuetify.VIcon("mdi-refresh", size=16)
                    vuetify.VTooltip("Refresh Contents", activator="parent")

            with GridLayout(columns=2, stretch=True):
                if show_directories:
                    with VBoxLayout(stretch=True):
                        vuetify.VListSubheader("Available Directories", classes="flex-0-1 justify-start pl-4")
                        with vuetify.VTreeview(
                            v_if=(f"{self._directories_name}.length > 0",),
                            activatable=True,
                            active_strategy="single-independent",
                            classes="flex-1-0 h-0 overflow-y-auto pt-0",
                            fluid=True,
                            item_value="path",
                            items=(self._directories_name,),
                            loading=True,
                            click_open=(self._vm.expand_directory, "[$event.path]"),
                            update_activated=(self.set_subdirectory, "$event"),
                        ):
                            with vuetify.Template(v_slot_title="{ item }"):
                                html.Span("{{ item.title }}", title=("item.title",))
                        vuetify.VListItem("No directories found", classes="flex-0-1 pl-4 text-left", v_else=True)

                with VBoxLayout(
                    classes="position-relative", column_span=1 if show_directories else 2, gap="0.5em", stretch=True
                ):
                    headers = kwargs.pop(
                        "headers",
                        ("[{ title: 'Available Datafiles', key: 'title' }]",),
                    )
                    super().__init__(
                        ref="test",
                        v_model=self._v_model,
                        classes="h-100 " + kwargs.pop("classes", ""),
                        headers=headers,
                        items=(self._datafiles_name,),
                        item_value="path",
                        select_strategy="all",
                        show_select=True,
                        raw_attrs=[
                            '''@click:row="(event, node) => node.toggleSelect(node.internalItem, node.index, event)"'''
                        ],
                        **kwargs,
                    )
                    with self:
                        with vuetify.Template(v_slot_no_data=True):
                            html.Span("No files to display.")
                    if self._label:
                        self.label = self._label
                    if "update_modelValue" not in kwargs:
                        self.update_modelValue = self._flush_state

            with InputField(
                v_show=self._show_selected_files.expression,
                v_model=self._v_model,
                classes="flex-0-1 nova-readonly",
                readonly=True,
                type="select",
                variant="outlined",
            ):
                with vuetify.Template(raw_attrs=['v-slot:selection="{ item, index }"']):
                    vuetify.VChip("{{ item.title.split('/').reverse()[0] }}", v_if="index < 2")
                    html.Span(
                        f"(+{{{{ {self._v_model}.length - 2 }}}} others)", v_if="index === 2", classes="text-caption"
                    )

                with vuetify.Template(v_slot_append_inner=True):
                    vuetify.VIcon(
                        "mdi-close-box", v_if=f"{self._v_model}.length > 0", color="primary", size=20, click=self.reset
                    )

    def _create_model(self) -> None:
        state = DataSelectorState()
        self._model = DataSelectorModel(state)

    def _create_viewmodel(self) -> None:
        server = get_server(None, client_type="vue3")
        binding = TrameBinding(server.state)

        self._vm = DataSelectorViewModel(self._model, binding)
        self._vm.state_bind.connect(self._state_name)
        self._vm.directories_bind.connect(self._directories_name)
        self._vm.datafiles_bind.connect(self._datafiles_name)
        self._vm.reset_bind.connect(self.reset)

    def refresh_contents(self) -> None:
        self._vm.update_view(refresh_directories=True)

    def reset(self, _: Any = None) -> None:
        if bool(get_state_param(self.state, self._clear_selection)):
            self._reset_state()

    def set_subdirectory(self, subdirectory_path: str = "") -> None:
        set_state_param(self.state, self._subdirectory, subdirectory_path)
        self._vm.set_subdirectory(subdirectory_path)

    def set_state(self, *args: Any, **kwargs: Any) -> None:
        raise TypeError(
            "The old DataSelector component has been renamed to NeutronDataSelector. Please import it from "
            "`nova.trame.view.components.ornl`."
        )

    # This method sets up Trame state change listeners for each binding parameter that can be changed directly by this
    # component. This allows us to communicate the changes to the developer's bindings without requiring our own. We
    # don't want bindings in the internal implementation as our callbacks could compete with the developer's.
    def _setup_bindings(self) -> None:
        # If the bindings were given initial values, write these to the state.
        set_state_param(self.state, self._directory)
        set_state_param(self.state, self._extensions)
        set_state_param(self.state, self._subdirectory)
        self._vm.set_binding_parameters(
            directory=get_state_param(self.state, self._directory),
            extensions=get_state_param(self.state, self._extensions),
            subdirectory=get_state_param(self.state, self._subdirectory),
        )

        # The component used by this parameter will attempt to set the initial value itself, which will trigger the
        # below change listeners causing unpredictable behavior.
        if isinstance(self._subdirectory, tuple):
            self._subdirectory = (self._subdirectory[0],)

        # Now we set up the change listeners for all bound parameters. These are responsible for updating the component
        # when other portions of the application manipulate these parameters.
        if isinstance(self._directory, tuple):

            @self.state.change(self._directory[0].split(".")[0])
            def on_directory_change(**kwargs: Any) -> None:
                directory = rgetdictvalue(kwargs, self._directory[0])
                if directory != self._last_directory:
                    self._last_directory = directory
                    self._vm.set_binding_parameters(
                        directory=set_state_param(self.state, self._directory, directory),
                    )

        if isinstance(self._extensions, tuple):

            @self.state.change(self._extensions[0].split(".")[0])
            def on_extensions_change(**kwargs: Any) -> None:
                extensions = rgetdictvalue(kwargs, self._extensions[0])
                if extensions != self._last_extensions:
                    self._last_extensions = extensions
                    self._vm.set_binding_parameters(
                        extensions=set_state_param(self.state, self._extensions, extensions),
                    )

        if isinstance(self._subdirectory, tuple):

            @self.state.change(self._subdirectory[0].split(".")[0])
            def on_subdirectory_change(**kwargs: Any) -> None:
                subdirectory = rgetdictvalue(kwargs, self._subdirectory[0])
                if subdirectory != self._last_subdirectory:
                    self._last_subdirectory = subdirectory
                    self._vm.set_binding_parameters(
                        subdirectory=set_state_param(self.state, self._subdirectory, subdirectory),
                    )

    async def _refresh_loop(self) -> None:
        refresh_rate: int = set_state_param(self.state, self._refresh_rate)
        skip = False

        if refresh_rate > 0:
            while True:
                await sleep(refresh_rate)
                if skip:
                    continue

                self.refresh_contents()
                self.state.dirty(self._datafiles_name)

                try:
                    refresh_rate = int(get_state_param(self.state, self._refresh_rate))
                    skip = False
                except TypeError:
                    refresh_rate = 1
                    skip = True



"""Module for the Tool outputs."""

from trame.app import get_server
from trame.widgets import vuetify3 as vuetify

from nova.mvvm.trame_binding import TrameBinding
from nova.trame.view.components import InputField
from nova.trame.view.layouts import HBoxLayout
from nova.trame.view_model.tool_outputs import ToolOutputsViewModel


class ToolOutputWindows:
    """Tool outputs class. Displays windows with tool stdout/stderr.

    This is intended to be used with the `nova-galaxy ToolRunner <https://nova-application-development.readthedocs.io/projects/nova-galaxy/en/latest/core_concepts/tool_runner.html>`__.
    """

    def __init__(self, id: str) -> None:
        """Constructor for ToolOutputWindows.

        Parameters
        ----------
        id : str
            Component id. Should be used consistently with ToolRunner and other components. Note that this parameter
            does not support Trame bindings.

        Returns
        -------
        None
        """
        self.id = f"tool_outputs_{id}"
        self.create_viewmodel(id)
        self.view_model.tool_outputs_bind.connect(self.id)
        self.create_ui()

    def create_viewmodel(self, id: str) -> None:
        server = get_server(None, client_type="vue3")
        binding = TrameBinding(server.state)
        self.view_model = ToolOutputsViewModel(id, binding)

    def create_ui(self) -> None:
        with HBoxLayout(stretch=True):
            with vuetify.VTabs(v_model=(f"{self.id}_active_output_tab", "1"), direction="vertical"):
                vuetify.VTab("Console output", value=1)
                vuetify.VTab("Console error", value=2)
            with HBoxLayout(stretch=True):
                InputField(
                    v_show=f"{self.id}_active_output_tab === '1'",
                    v_model=f"{self.id}.stdout",
                    id=f"{self.id}_outputs",
                    type="autoscroll",
                    no_resize=True,
                    readonly=True,
                )
                InputField(
                    v_show=f"{self.id}_active_output_tab === '2'",
                    v_model=f"{self.id}.stderr",
                    id=f"{self.id}_errors",
                    type="autoscroll",
                    no_resize=True,
                    readonly=True,
                )



"""View implementation for MatplotlibFigure."""

import json
import os
import socketserver
from asyncio import FIRST_COMPLETED, create_task, new_event_loop, set_event_loop, wait
from io import BytesIO
from mimetypes import types_map
from pathlib import Path
from threading import Thread
from typing import Any, Optional
from warnings import warn

import tornado
from aiohttp import ClientSession, WSMsgType, web
from matplotlib import get_data_path, rcParams
from matplotlib.backends.backend_webagg import FigureManagerWebAgg, new_figure_manager_given_figure  # type: ignore
from matplotlib.figure import Figure
from trame.app import get_server
from trame.widgets import client, html, matplotlib
from wslink.backends.aiohttp import WebAppServer


class _MPLApplication(tornado.web.Application):
    """Tornado application compatible with Matplotlib's WebAgg backend."""

    class WebSocket(tornado.websocket.WebSocketHandler):
        """Implements the WebSocket manager for WebAgg."""

        supports_binary = True

        def check_origin(self, origin: Any) -> bool:
            return True

        def open(self, *args: Any, **kwargs: Any) -> None:
            # Register the websocket with the FigureManager.
            manager = self.application.manager  # type: ignore
            manager.add_web_socket(self)
            if hasattr(self, "set_nodelay"):
                self.set_nodelay(True)

        def on_close(self) -> None:
            # When the socket is closed, deregister the websocket with
            # the FigureManager.
            manager = self.application.manager  # type: ignore
            manager.remove_web_socket(self)

        def on_message(self, message: Any) -> None:
            # The 'supports_binary' message is relevant to the
            # websocket itself.  The other messages get passed along
            # to matplotlib as-is.

            # Every message has a "type" and a "figure_id".
            message = json.loads(message)
            if message["type"] == "supports_binary":
                self.supports_binary = message["value"]
            else:
                manager = self.application.manager  # type: ignore
                try:
                    manager.handle_json(message)
                except Exception:
                    manager.refresh_all()

        def send_json(self, content: Any) -> None:
            set_event_loop(self.application.loop)  # type: ignore
            self.write_message(json.dumps(content))

        def send_binary(self, blob: Any) -> None:
            set_event_loop(self.application.loop)  # type: ignore
            if self.supports_binary:
                self.write_message(blob, binary=True)
            else:
                data_uri = "data:image/png;base64," + blob.encode("base64").replace("\n", "")
                self.write_message(data_uri)

        def write_message(self, *args: Any, **kwargs: Any) -> Any:
            # We need the websocket to remain alive if a message fails to write.
            try:
                super().write_message(*args, **kwargs)
            except Exception:
                pass

    def __init__(self, figure: Figure) -> None:
        self.figure = figure
        self.manager = new_figure_manager_given_figure(id(figure), figure)
        self.loop = new_event_loop()

        super().__init__([("/ws", self.WebSocket)])


class MatplotlibFigure(matplotlib.Figure):
    """Creates an interactive Matplotlib Figure using the WebAgg backend.

    By default, this will leverage the built-in Trame widget for Matplotlib support. This built-in widget can display
    poor performance for detailed plots due to it being locked into using SVG rendering.

    If you experience this, then you can use the `webagg` parameter to enable the WebAgg backend for Matplotlib. This
    will switch to server-side rendering leveraging the Anti-Grain Geometry engine.

    .. code-block:: python

        my_figure = matplotlib.figure.Figure()
        MatplotlibFigure(my_figure)  # Display SVG-based plot in Trame
        MatplotlibFigure(my_figure, webagg=True)  # Display WebAgg-based plot in Trame
    """

    mpl_initialized = False
    mpl_instances: dict[int, "MatplotlibFigure"] = {}

    @classmethod
    def _get_free_port(cls) -> int:
        with socketserver.TCPServer(("localhost", 0), None) as s:  # type: ignore
            return s.server_address[1]

    @classmethod
    def _setup_mpl(cls) -> None:
        server = get_server(None, client_type="vue3")

        @server.controller.add("on_server_bind")
        def _add_routes(wslink_server: WebAppServer) -> None:
            # matplotlib WebAgg serves JS that will reference the base URL of the application, so we need to add
            # endpoints to the main server to handle these requests.
            wslink_server.app.add_routes(
                [
                    web.get("/_images/{image}", cls._mpl_image_endpoint),
                    web.get("/download/{port}/{format}", cls._mpl_download_endpoint),
                    web.get("/mpl/{port}", cls._mpl_ws_proxy_endpoint),
                ]
            )

        # The CSS and JS files, on the other hand, can be preloaded into the page which is simpler.
        css_path = Path(FigureManagerWebAgg.get_static_file_path(), "css")
        for fname in os.listdir(css_path):
            with open(Path(css_path, fname)) as css_file:
                content = css_file.read()
                client.Style(content)
        js = FigureManagerWebAgg.get_javascript()
        client.Script(js.replace("window.setTimeout(set_focus, 100);", "//"))

        MatplotlibFigure.mpl_initialized = True

    @classmethod
    async def _mpl_download_endpoint(cls, request: web.Request) -> web.Response:
        # We use the websocket port to differentiate between plots if there are multiple WebAgg figures.
        port = request.match_info.get("port", "")
        format = request.match_info.get("format", "png")

        if not port or int(port) not in MatplotlibFigure.mpl_instances:
            raise web.HTTPNotFound

        instance = MatplotlibFigure.mpl_instances[int(port)]

        buff = BytesIO()
        if instance._figure:
            instance._figure.savefig(buff, format=format)
        buff.seek(0)

        return web.Response(body=buff.read(), headers={"Content-Type": types_map.get(format, "binary")})

    @classmethod
    async def _mpl_image_endpoint(cls, request: web.Request) -> web.Response:
        image_name = request.match_info.get("image", None)

        if image_name:
            try:
                with open(Path(get_data_path(), "images", image_name), mode="rb") as image_file:
                    image_data = image_file.read()
                    return web.Response(body=image_data)
            except OSError as err:
                raise web.HTTPNotFound from err

        raise web.HTTPNotFound

    @classmethod
    async def _mpl_ws_proxy_endpoint(cls, request: web.Request) -> web.WebSocketResponse:
        # The WebAgg backend assumes a tornado-based WebSocket handler, so we need to proxy it to work with Trame's
        # aiohttp setup.
        port = request.match_info.get("port", "")

        # Initialize the proxy
        ws_server = web.WebSocketResponse()
        await ws_server.prepare(request)

        # Connect to the tornado WebSocket handler
        client_session = ClientSession(cookies=request.cookies)
        async with client_session.ws_connect(
            f"http://localhost:{port}/ws",
        ) as ws_client:

            async def ws_forward(ws_from: Any, ws_to: Any) -> None:
                # The browser will send text messages for rendering requests and the server will send bytes to transmit
                # rendered image data.
                async for msg in ws_from:
                    if msg.type == WSMsgType.TEXT:
                        await ws_to.send_str(msg.data)
                    elif msg.type == WSMsgType.BINARY:
                        await ws_to.send_bytes(msg.data)
                    else:
                        raise ValueError("unexpected message type: %s", print(msg))

            # Forward websocket data in both directions
            server_to_client = create_task(ws_forward(ws_server, ws_client))
            client_to_server = create_task(ws_forward(ws_client, ws_server))
            await wait([server_to_client, client_to_server], return_when=FIRST_COMPLETED)
            await client_session.close()  # Ensure the connection is cleaned up when the Trame client disconnects.

            return ws_server

    def __init__(self, figure: Optional[Figure] = None, webagg: bool = False, **kwargs: Any) -> None:
        """Creates a Matplotlib figure in the Trame UI.

        Parameters
        ----------
        figure : `matplotlib.figure.Figure <https://matplotlib.org/stable/api/_as_gen/matplotlib.figure.Figure.html#matplotlib.figure.Figure>`__, optional
            Initial Matplotlib figure.
        webagg : bool, optional
            If true, then the WebAgg backend for Matplotlib is used. If not, then the default Trame matplotlib plugin
            is used. Note that this parameter does not supporting Trame bindings since the user experiences are
            fundamentally different between the two options and toggling them is not considered a good idea by the
            author of this component.
        kwargs
            Arguments to be passed to `AbstractElement <https://trame.readthedocs.io/en/latest/core.widget.html#trame_client.widgets.core.AbstractElement>`_

        Returns
        -------
        None
        """  # noqa: E501
        self._server = get_server(None, client_type="vue3")
        self._webagg = webagg
        if "classes" in kwargs:
            kwargs["classes"] += " flex-1-1"
        else:
            kwargs["classes"] = "flex-1-1"
        if webagg:
            if "id" in kwargs:
                kwargs.pop("id")
                warn("id parameter to MatplotlibFigure is ignored when webagg=True.", stacklevel=1)

            self._port = MatplotlibFigure._get_free_port()
            self._id = f"nova_mpl_{self._port}"
            kwargs["classes"] += " nova-mpl"

            html.Div(id=self._id, **kwargs)

            self._figure = figure
            self._initialized = False

            if not MatplotlibFigure.mpl_initialized:
                MatplotlibFigure._setup_mpl()
            MatplotlibFigure.mpl_instances[self._port] = self

            self.update()
        else:
            super().__init__(figure, **kwargs)
            self._id = self._key

        self._query_selector = f"window.document.querySelector('#{self._id}')"
        self._trigger = (
            f"if ({self._query_selector} === null) {{ return; }}"
            # webagg figures receive a fixed width and height. This blocks the flexbox scaling, so I temporarily hide
            # the figure to allow the container to grow/shrink naturally in flexbox.
            f"window.document.querySelectorAll('.nova-mpl').forEach((item) => {{ item.style.display = 'none'; }});"
            f"const height = {self._query_selector}.parentNode.offsetHeight;"
            f"const width = {self._query_selector}.parentNode.offsetWidth;"
            # Revert the display value to allow the figure to render again.
            f"window.document.querySelectorAll('.nova-mpl').forEach((item) => {{ item.style.display = ''; }});"
            "window.trame.trigger("
            f"  '{self._id}_resize',"
            f"  [height, width]"
            ");"
        )
        self._resize_figure = client.JSEval(exec=self._trigger).exec
        self._resize_listener = client.JSEval(
            exec=(
                # ResizeObserver is necessary to detect changes in size unrelated to the viewport size such as when
                # content is conditionally rendered that changes the size of the figure's container.
                "const resizeObserver = new window.ResizeObserver(() => {"
                f"  window.delay_manager.debounce('{self._id}', function() {{ {self._trigger} }}, 500);"
                "});"
                f"resizeObserver.observe({self._query_selector}.parentNode);"
            )
        ).exec

        @self._server.controller.trigger(f"{self._id}_resize")
        def resize_figure(height: int, width: int) -> None:
            if self._figure:
                if self._webagg:
                    # Reserve space for the controls injected by webagg.
                    height -= 48
                    width -= 4

                if height <= 0 or width <= 0:
                    return

                if self._webagg:
                    # Webagg does not respect the Figure object's DPI.
                    dpi = rcParams["figure.dpi"]
                else:
                    dpi = self._figure.get_dpi()
                new_width = width / dpi
                new_height = height / dpi
                current_size = self._figure.get_size_inches()
                if current_size[0] != new_width or current_size[1] != new_height:
                    self._figure.set_size_inches(new_width, new_height)

                self.update(skip_resize=True)

        client.ClientTriggers(mounted=self._resize_listener)
        client.ClientTriggers(mounted=self._resize_figure)

    def update(self, figure: Optional[Figure] = None, skip_resize: bool = False) -> None:
        if self._webagg:
            if figure:
                self._figure = figure

            if self._figure is not None and not self._initialized:
                self._setup_figure_websocket()
                self._setup_figure_javascript()

                self._initialized = True

            # Re-render the figure in the UI
            if self._figure is not None:
                self._figure.canvas.draw_idle()
                self._figure.canvas.flush_events()
        else:
            super().update(figure)

        if not skip_resize and hasattr(self, "_resize_figure"):
            self._resize_figure()
        else:
            self._server.state.flush()

    def _setup_figure_websocket(self) -> None:
        thread = Thread(target=self._mpl_run_ws_server, daemon=True)
        thread.start()

    def _setup_figure_javascript(self) -> None:
        figure_js = """
            function ondownload_%(port)d(figure, format) {
                window.open('download/%(port)d/' + format, '_blank');
            };

            function ready_%(port)d(fn) {
                if (document.getElementById("nova_mpl_%(port)d") === null) {
                    setTimeout(() => {
                        ready_%(port)d(fn);
                    }, 100);
                } else {
                    fn();
                }
            }

            ready_%(port)d(
                function() {
                    var websocket_type = mpl.get_websocket_type();
                    var websocket = new websocket_type('mpl/%(port)d');

                    var fig = new mpl.figure(
                        // A unique numeric identifier for the figure
                        %(port)d,
                        // A websocket object (or something that behaves like one)
                        websocket,
                        // A function called when a file type is selected for download
                        ondownload_%(port)d,
                        // The HTML element in which to place the figure
                        document.getElementById("nova_mpl_%(port)d")
                    );
                }
            );
        """

        client.Script(figure_js % {"port": self._port})  # TODO

    def _mpl_run_ws_server(self) -> None:
        if not self._figure:
            return

        application = _MPLApplication(self._figure)

        http_server = tornado.httpserver.HTTPServer(application)
        sockets = tornado.netutil.bind_sockets(self._port, "")
        http_server.add_sockets(sockets)

        ioloop = tornado.ioloop.IOLoop.instance()
        ioloop.start()



"""Model for MVVM demo of NeutronDataSelector."""

from typing import List

from pydantic import BaseModel, Field


class NeutronDataSelectorBindingTest(BaseModel):
    """Model for testing binding parameters for NeutronDataSelector."""

    facility: str = Field(default="", title="Facility")
    instrument: str = Field(default="", title="Instrument")
    experiment: str = Field(default="", title="Experiment")
    allow_custom_directories: bool = Field(default=False, title="Allow Custom Directories?")


class AnalysisDataSelectorState(BaseModel):
    """Model for MVVM demo of NeutronDataSelector."""

    selected_files: List[str] = Field(default=[], title="Selected Files")


class ONCatDataSelectorState(BaseModel):
    """Model for MVVM demo of NeutronDataSelector."""

    selected_files: List[str] = Field(default=[], title="Selected Files")



"""Module for the JobProgress ViewModel."""

import asyncio
from typing import Any, Optional

import blinker
from pydantic import BaseModel

from nova.common.job import WorkState
from nova.common.signals import Signal, ToolCommand, get_signal_id
from nova.mvvm.interface import BindingInterface


def job_running(status: WorkState) -> bool:
    """A helper function to check if job is doing something in Galaxy."""
    return status in [
        WorkState.UPLOADING_DATA,
        WorkState.QUEUED,
        WorkState.RUNNING,
        WorkState.STOPPING,
        WorkState.CANCELING,
    ]


class ButtonsState(BaseModel):
    """Class that manages start/stop/cancel button states."""

    run_disabled: bool = False
    cancel_disabled: bool = True
    stop_disabled: bool = True
    download_disabled: bool = True

    stop_in_progress: bool = False
    cancel_in_progress: bool = False
    download_in_progress: bool = False

    def update_from_workstate(self, status: WorkState) -> None:
        running = job_running(status)
        self.run_disabled = running
        self.cancel_disabled = not running
        self.stop_disabled = status not in [WorkState.RUNNING, WorkState.STOPPING]
        self.stop_in_progress = status == WorkState.STOPPING
        self.cancel_in_progress = status == WorkState.CANCELING
        self.download_disabled = status != WorkState.FINISHED


class ExecutionButtonsViewModel:
    """A viewmodel responsible for execution buttons."""

    def __init__(self, id: str, binding: BindingInterface):
        self.sender_id = f"ExecutionButtonsViewModel_{id}"
        self.button_states = ButtonsState()
        self.buttons_state_bind = binding.new_bind(self.button_states)
        self.execution_signal = blinker.signal(get_signal_id(id, Signal.TOOL_COMMAND))
        self.progress_signal = blinker.signal(get_signal_id(id, Signal.PROGRESS))
        self.progress_signal.connect(self.update_state, weak=False)

    async def update_state(self, _sender: Any, state: WorkState, details: str) -> None:
        self.button_states.update_from_workstate(state)
        self.buttons_state_bind.update_in_view(self.button_states)

    async def run(self) -> None:
        # disable run now since it might take some time before the client updates the status
        self.button_states.run_disabled = True
        self.buttons_state_bind.update_in_view(self.button_states)
        await self.execution_signal.send_async(self.sender_id, command=ToolCommand.START)

    async def cancel(self) -> None:
        await self.execution_signal.send_async(self.sender_id, command=ToolCommand.CANCEL)

    async def stop(self) -> None:
        await self.execution_signal.send_async(self.sender_id, command=ToolCommand.CANCEL)

    async def prepare_results(self) -> Optional[bytes]:
        self.button_states.download_in_progress = True
        self.buttons_state_bind.update_in_view(self.button_states)
        await asyncio.sleep(0.5)  # to give Trame time to update view
        responses = await self.execution_signal.send_async(self.sender_id, command=ToolCommand.GET_RESULTS)
        res = None
        for response in responses:  # responses can come from multiple places
            if response[1] is None:
                continue
            if response[1]["sender"] == self.sender_id and response[1]["command"] == ToolCommand.GET_RESULTS:
                res = response[1]["results"]
        self.button_states.download_in_progress = False
        self.buttons_state_bind.update_in_view(self.button_states)
        return res



"""Entrypoint for the widget gallery."""

import sys
from typing import Any

from trame_server.core import Server

from tests.gallery import App


def main(server: Server = None, **kwargs: Any) -> None:
    # [run app]
    app = App(server)
    for arg in sys.argv[1:]:
        try:
            key, value = arg.split("=")
            kwargs[key] = int(value)
        except Exception:
            pass
    app.server.start(**kwargs)
    # [run app complete]



"""View model implementation for the DataSelector widget."""

import os
from typing import Any, Dict, List

from nova.mvvm.interface import BindingInterface
from nova.trame.model.ornl.neutron_data_selector import NeutronDataSelectorModel
from nova.trame.view_model.data_selector import DataSelectorViewModel


class NeutronDataSelectorViewModel(DataSelectorViewModel):
    """Manages the view state of the DataSelector widget."""

    def __init__(self, model: NeutronDataSelectorModel, binding: BindingInterface) -> None:
        super().__init__(model, binding)
        self.model: NeutronDataSelectorModel = model

        self.facilities_bind = binding.new_bind()
        self.instruments_bind = binding.new_bind()
        self.experiments_bind = binding.new_bind()

    def reset(self) -> None:
        self.model.set_subdirectory("")
        self.directories = self.model.get_directories()
        self.expanded = {}

        self.update_view()
        self.reset_bind.update_in_view(None)

    def on_state_updated(self, results: Dict[str, Any]) -> None:
        for update in results.get("updated", []):
            match update:
                case "custom_directory":
                    self.reset()
                    self.update_view()
                case "search":
                    self.update_view()

    def transform_datafiles(self, datafiles: List[Any]) -> List[Dict[str, str]]:
        return [{"title": os.path.basename(datafile["path"]), **datafile} for datafile in datafiles]

    def update_view(self, refresh_directories: bool = False) -> None:
        self.facilities_bind.update_in_view(self.model.get_facilities())
        self.instruments_bind.update_in_view(self.model.get_instruments())
        self.experiments_bind.update_in_view(self.model.get_experiments())

        super().update_view(refresh_directories)



"""Unit tests for FileUpload."""

from trame.app import get_server
from trame_server.core import Server

from nova.trame.view.components import FileUpload
from nova.trame.view.theme import ThemedApp


def test_file_upload() -> None:
    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            server = get_server(None, client_type="vue3")
            super().__init__(server=server)
            self.create_ui()

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    file_upload = FileUpload(v_model="test", base_paths=["/HFIR", "/SNS"], label="Test Upload")
                    assert file_upload.remote_file_input.v_model == "test"
                    assert file_upload.remote_file_input.base_paths == ["/HFIR", "/SNS"]
                    assert file_upload.children[0] == "Test Upload"

    MyTrameApp()



"""Trame implementation of the VBoxLayout class."""

from typing import Any, Optional, Union

from trame.widgets import html

from .utils import merge_styles


class VBoxLayout(html.Div):
    """Creates an element that vertically stacks its children."""

    def __init__(
        self,
        height: Optional[Union[int, str]] = None,
        width: Optional[Union[int, str]] = None,
        halign: Optional[str] = None,
        valign: Optional[str] = None,
        gap: Optional[Union[int, str]] = "0em",
        vspace: Optional[Union[int, str]] = "0em",
        stretch: bool = False,
        **kwargs: Any,
    ) -> None:
        """Constructor for VBoxLayout.

        Parameters
        ----------
        height : optional[int | str]
            The height of this box. If an integer is provided, it is interpreted as pixels. If a string is provided,
            the string is treated as a CSS value.
        width : optional[int | str]
            The width of this box. If an integer is provided, it is interpreted as pixels. If a string is provided,
            the string is treated as a CSS value.
        halign : optional[str]
            The horizontal alignment of items in the grid. See `MDN
            <https://developer.mozilla.org/en-US/docs/Web/CSS/align-items>`__ for available options.
        valign : optional[str]
            The vertical alignment of items in the grid. See `MDN
            <https://developer.mozilla.org/en-US/docs/Web/CSS/justify-items>`__ for available options.
        gap : optional[str]
            The horizontal gap to place between items. Can be any CSS gap value (e.g. "4px" or "0.25em"). Defaults to no
            gap between items.
        vspace : optional[str]
            The vertical gap to place between items. Can be any CSS gap value (e.g. "4px" or "0.25em"). Defaults to no
            gap between items.
        stretch : optional[bool]
            If True, then this layout component will stretch to attempt to fill the space of it's parent container.
            Defaults to False.
        kwargs : Any
            Additional keyword arguments to pass to html.Div.

        Returns
        -------
        None

        Example
        -------
        .. literalinclude:: ../tests/gallery/views/app.py
            :start-after: setup vbox
            :end-before: setup vbox complete
            :dedent:
        """
        classes = kwargs.pop("classes", [])
        if isinstance(classes, list):
            classes = " ".join(classes)

        if stretch:
            classes += " flex-1-1 overflow-y-auto"
        else:
            classes += " flex-0-1"

        v_show = kwargs.get("v_show", None)
        if v_show:
            classes = (f"{v_show} ? '{classes} d-flex flex-column' : '{classes}'",)
        else:
            classes += " d-flex flex-column"

        widget_style = self.get_root_styles(height, width, halign, valign, gap, vspace)
        user_style = kwargs.pop("style", {})

        super().__init__(classes=classes, style=merge_styles(widget_style, user_style), **kwargs)

    def get_root_styles(
        self,
        height: Optional[Union[int, str]],
        width: Optional[Union[int, str]],
        halign: Optional[str],
        valign: Optional[str],
        gap: Optional[Union[int, str]],
        vspace: Optional[Union[int, str]],
    ) -> dict:
        height = f"{height}px" if isinstance(height, int) else height
        width = f"{width}px" if isinstance(width, int) else width
        gap = f"{gap}px" if isinstance(gap, int) else gap
        vspace = f"{vspace}px" if isinstance(vspace, int) else vspace

        styles = {}

        if height:
            styles["height"] = height
        if width:
            styles["width"] = width
        if halign:
            styles["align-items"] = halign
        if valign:
            styles["justify-content"] = valign
        if gap:
            styles["gap"] = gap
        if vspace:
            styles["margin-bottom"] = vspace

        return styles



"""Module for the Progress Tab."""

from trame.app import get_server
from trame.widgets import vuetify3 as vuetify
from trame_client.widgets import html

from nova.mvvm.trame_binding import TrameBinding
from nova.trame.view.layouts import VBoxLayout
from nova.trame.view_model.progress_bar import ProgressBarViewModel


class ProgressBar:
    """Progress bar class. Adds progress bar that displays job status to the view.

    This is intended to be used with the `nova-galaxy ToolRunner <https://nova-application-development.readthedocs.io/projects/nova-galaxy/en/latest/core_concepts/tool_runner.html>`__.
    """

    def __init__(self, id: str) -> None:
        """Constructor for ProgressBar.

        Parameters
        ----------
        id : str
            Component id. Should be used consistently with ToolRunner and other components. Note that this parameter
            does not support Trame bindings.

        Returns
        -------
        None
        """
        self.id = f"progress_bar_{id}"
        self.create_viewmodel(id)
        self.view_model.progress_state_bind.connect(self.id)
        self.create_ui()

    def create_viewmodel(self, id: str) -> None:
        server = get_server(None, client_type="vue3")
        binding = TrameBinding(server.state)
        self.view_model = ProgressBarViewModel(id, binding)

    def create_ui(self) -> None:
        with VBoxLayout(v_show=f"{self.id}.show_progress || {self.id}.show_ok || {self.id}.show_failed", height=25):
            with vuetify.VProgressLinear(
                height="25",
                model_value=(f"{self.id}.progress", "0"),
                striped=True,
                id=f"{self.id}_show_progress",
                v_show=(f"{self.id}.show_progress",),
            ):
                html.H5(v_text=f"{self.id}.details")
                with vuetify.VMenu(
                    max_width=900,
                    location="bottom",
                    no_click_animation=True,
                    close_on_content_click=False,
                    open_on_hover=True,
                    v_show=False,
                ):
                    with vuetify.Template(v_slot_activator="{ props }"):
                        vuetify.VIcon(
                            "mdi-information",
                            v_show=f"{self.id}.show_full_details",
                            v_bind="props",
                            classes="ml-2",
                            color="primary",
                        )

                    with vuetify.VCard(classes="bg-grey"):
                        vuetify.VCardText(f"{{{{ {self.id}.full_details }}}}", classes="display-linebreaks")

            with vuetify.VProgressLinear(
                height="25",
                model_value="100",
                striped=False,
                color="error",
                id=f"{self.id}_show_failed",
                v_show=(f"{self.id}.show_failed",),
            ):
                html.H5(v_text=f"{self.id}.details", classes="text-white")
            with vuetify.VProgressLinear(
                height="25",
                model_value="100",
                striped=False,
                color="primary",
                id=f"{self.id}_show_ok",
                v_show=(f"{self.id}.show_ok",),
            ):
                html.H5(v_text=f"{self.id}.details", classes="text-white")



"""ONCat backend for NeutronDataSelector."""

import os
from pathlib import Path
from typing import Any, Dict, List, Optional

from natsort import natsorted
from pydantic import Field
from pyoncat import CLIENT_CREDENTIALS_FLOW, ONCat

from .neutron_data_selector import NeutronDataSelectorModel, NeutronDataSelectorState

TOKEN_VARNAME = "USER_OIDC_TOKEN"
ID_VARNAME = "ONCAT_CLIENT_ID"
SECRET_VARNAME = "ONCAT_CLIENT_SECRET"


class ONCatDataSelectorState(NeutronDataSelectorState):
    """Selection state for identifying datafiles."""

    instrument_mapping: Dict[str, str] = Field(default={})
    projection: List[str] = Field(default=[])


class ONCatDataSelectorModel(NeutronDataSelectorModel):
    """ONCat backend for NeutronDataSelector."""

    def __init__(self, state: ONCatDataSelectorState) -> None:
        super().__init__(state)
        self.state: ONCatDataSelectorState = state

        user_token = os.environ.get(TOKEN_VARNAME, "")
        client_id = os.environ.get(ID_VARNAME, "")
        client_secret = os.environ.get(SECRET_VARNAME, "")
        if user_token:
            self.oncat_client = ONCat(url="https://calvera-test.ornl.gov/oncat", api_token=user_token)
        elif client_id and client_secret:
            self.oncat_client = ONCat(
                url="https://oncat.ornl.gov",
                client_id=client_id,
                client_secret=client_secret,
                flow=CLIENT_CREDENTIALS_FLOW,
            )
        else:
            raise EnvironmentError(
                f"In order to use the ONCat backend for NeutronDataSelector, you must set either {TOKEN_VARNAME} or "
                f"both {ID_VARNAME} and {SECRET_VARNAME} in your environment."
            )

    def set_binding_parameters(self, **kwargs: Any) -> None:
        super().set_binding_parameters(**kwargs)

        if "projection" in kwargs:
            self.state.projection = kwargs["projection"]

    def get_facilities(self) -> List[str]:
        facilities = []
        for facility_data in self.oncat_client.Facility.list(projection=["name"]):
            facilities.append(facility_data.name)
        return natsorted(facilities)

    def get_instruments(self) -> List[Dict[str, str]]:
        if not self.state.facility:
            return []

        self.state.instrument_mapping = {}
        instruments = []
        for instrument_data in self.oncat_client.Instrument.list(
            facility=self.state.facility, projection=["long_id", "short_name"]
        ):
            self.state.instrument_mapping[instrument_data.short_name] = instrument_data.id

            id = instrument_data.long_id
            name = instrument_data.short_name
            instruments.append({"id": id, "name": name, "title": f"{id}: {name}"})

        return natsorted(instruments, key=lambda x: x["name"])

    def get_experiments(self) -> List[str]:
        if not self.state.facility or not self.state.instrument:
            return []

        experiments = []
        for experiment_data in self.oncat_client.Experiment.list(
            facility=self.state.facility,
            instrument=self.state.instrument_mapping[self.state.instrument],
            projection=["name"],
        ):
            experiments.append(experiment_data.name)
        return natsorted(experiments)

    def get_directories(self, _: Optional[Path] = None) -> List[Dict[str, Any]]:
        return []

    def create_datafile_obj(self, data: Dict[str, Any], projection: List[str]) -> Dict[str, str]:
        new_obj = {"path": data["location"]}

        for key in projection:
            value: Any = data

            if key == "location":
                continue

            for part in key.split("."):
                try:
                    value = value[part]
                except KeyError:
                    value = ""
                    break

            new_obj[key] = value

        return new_obj

    def get_datafiles(self, *args: Any, **kwargs: Any) -> List[Dict[str, str]]:
        if not self.state.facility or not self.state.instrument or not self.state.experiment:
            return []

        projection = ["location"] + self.state.projection

        datafiles = []
        for datafile_data in self.oncat_client.Datafile.list(
            facility=self.state.facility,
            instrument=self.state.instrument_mapping[self.state.instrument],
            experiment=self.state.experiment,
            projection=projection,
        ):
            can_add = False
            path = datafile_data.location
            if self.state.extensions:
                for extension in self.state.extensions:
                    if path.lower().endswith(extension):
                        can_add = True
            else:
                can_add = True

            if self.state.search and self.state.search.lower() not in os.path.basename(path).lower():
                can_add = False

            if can_add:
                datafiles.append(self.create_datafile_obj(datafile_data, projection))

        return natsorted(datafiles, key=lambda d: d["path"])



"""Model implementation for DataSelector."""

from pathlib import Path
from typing import Any, Dict, List, Optional

from natsort import natsorted
from pydantic import Field, field_validator

from ..data_selector import DataSelectorModel, DataSelectorState


class NeutronDataSelectorState(DataSelectorState):
    """Selection state for identifying datafiles."""

    facility: str = Field(default="", title="Facility")
    instrument: str = Field(default="", title="Instrument")
    experiment: str = Field(default="", title="Experiment")

    @field_validator("experiment", mode="after")
    @classmethod
    def validate_experiment(cls, experiment: str) -> str:
        if experiment and not experiment.startswith("IPTS-"):
            raise ValueError("experiment must begin with IPTS-")
        return experiment

    def get_facilities(self) -> List[str]:
        raise NotImplementedError()

    def get_instruments(self) -> List[Dict[str, str]]:
        raise NotImplementedError()


class NeutronDataSelectorModel(DataSelectorModel):
    """Manages file system interactions for the DataSelector widget."""

    def __init__(self, state: NeutronDataSelectorState) -> None:
        super().__init__(state)
        self.state: NeutronDataSelectorState = state

    def set_binding_parameters(self, **kwargs: Any) -> None:
        super().set_binding_parameters(**kwargs)

        if "facility" in kwargs:
            self.state.facility = kwargs["facility"]
        if "instrument" in kwargs:
            self.state.instrument = kwargs["instrument"]
        if "experiment" in kwargs:
            self.state.experiment = kwargs["experiment"]

    def get_facilities(self) -> List[str]:
        return natsorted(self.state.get_facilities())

    def get_instruments(self) -> List[Dict[str, str]]:
        return natsorted(self.state.get_instruments())

    def get_experiments(self) -> List[str]:
        raise NotImplementedError()

    def get_directories(self, base_path: Optional[Path] = None) -> List[Dict[str, Any]]:
        raise NotImplementedError()

    def get_datafiles(self, *args: Any, **kwargs: Any) -> List[Dict[str, str]]:
        raise NotImplementedError()
